#!/usr/bin/perl -w
use strict;
#use diagnostics;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

require cgi;
require db;
require cvsweb;
require cvsdb;
require error_html;

#  This file contains the code to *edit* access lists.  The code to actually
#  use them gets used all the time, and we want to load it without this rare
#  edit stuff, so we put the use code in session.pm

package access_html;

my %privHelp = (
     0x0004 => [ "visible", "Can see file exists"],
     0x0008 => [ "read-f",  "Can read file text/diffs (this branch)"],
     0x0020 => [ "delete-f","Can delete file (all revisions)"],
     0x0010 => [ "commit",  "Can check-in this branch"],
     0x0020 => [ "read-c",  "Can read comment"],
     0x0040 => [ "edit-c",  "Can edit the comment"],
     0x0080 => [ "append-c","Can append to the comment"],
     0x0100 => [ "edit-p",  "Can change access permissions"],
     0x0200 => [ "read-p",  "Can read access permissions"],
     0x0400 => [ "create-b","Can create a branch"],
     0x0800 => [ "delete-b","Can delete branch"],
     0x1000 => [ "create-l","Can create label"],
     0x2000 => [ "delete-l","Can delete a label"],
     0x4000 => [ "move-l",  "Can move a label"],
);

#  We are going to be permissive.  The first three tokens can be in any order.
#  The only one which is not user|group|allow|deny will be the who.
sub textToAcl {
   my($text) = (@_);

   my $acl = "";
   my $lineNum = 0;
   foreach my $line (split(/^/m,$text)) {  # for each line
      $lineNum++;  # for prettier error messages
      $line =~ s/^\s+//;  # delete leading white space
      my($a,$b,$c,$privs) = split(/\s+/,$line,4);
      my($usergroup,$whotext,$allowdeny);
      next if !defined $a || $a eq "";  # skip blank line
      if(!defined $privs || $privs eq "") {
         error_html::fatal("line $lineNum: too short");
         exit;
      };
      #print(STDERR "a={$usergroup} b={$whotext} c={$allowdeny} priv={$privs}\n");
      foreach my $token ($a,$b,$c) {
         if($token =~ /^user$/i) {
            if(defined $usergroup) {
               error_html::fatal("line $lineNum: syntax error at <tt>". cgi::htmlQuote($token)."</tt>");
               exit;
            };
            $usergroup = 2;  # user
         } elsif($token =~ /^group$/i) {
            if(defined $usergroup) {
               error_html::fatal("line $lineNum: syntax error at <tt>". cgi::htmlQuote($token)."</tt>");
               exit;
            };
            $usergroup = 0;
         } elsif($token =~ /^allow$/i) {
            if(defined $allowdeny) {
               error_html::fatal("line $lineNum: syntax error at <tt>". cgi::htmlQuote($token)."</tt>");
               exit;
            };
            $allowdeny = 1;   # allow
         } elsif($token =~ /^deny$/i) {
            if(defined $allowdeny) {
               error_html::fatal("line $lineNum: syntax error at <tt>". cgi::htmlQuote($token)."</tt>");
               exit;
            };
            $allowdeny = 0;
         } else {  # else this must be the name of the user or group
            if(defined $whotext) {
               error_html::fatal("line $lineNum: syntax error at <tt>". cgi::htmlQuote($token)."</tt>");
               exit;
            };
            $whotext = $token;
         }
      };
      my $who;
      if($whotext =~ /^\d+$/) {   # if $whotext is all numeric
         $who = $whotext;
      } elsif($usergroup) {
         db::SendSQL("SELECT userid FROM user WHERE username = ".db::SqlQuote($whotext));
         $who = db::FetchSQLData();
      } else {
         db::SendSQL("SELECT group_id FROM groups WHERE name = ".db::SqlQuote($whotext));
         $who = db::FetchSQLData();
      };
      if(! defined $who) {
         error_html::fatal("line $lineNum: I don't know user/group <tt>". cgi::htmlQuote($whotext)."</tt>");
         exit;
      }
      my $flags = ($usergroup | $allowdeny);
      my %privNum;  # key is priv-name value priv-bit
      foreach my $bit (keys %privHelp) {  
         $privNum{${$privHelp{$bit}}[0]} = $bit; 
      };
      foreach my $priv (split(/[^\w-]+/sm,$privs)) {  # for each alpha-numeric string (punctuation agnostic)
         next if $priv eq "";  # skip empty privs
         (my $lpriv = $priv) =~ tr/A-Z/a-z/;  # make lower case
         if(defined $privNum{$lpriv}) {  # if is legit priv
            $flags |= $privNum{$lpriv};
         } elsif($lpriv =~ /^0x([0-9a-f]+)$/) {  # else priv looks like hex number
            $flags |= hex($1) & 0xfffc;  # mask off group/user and allow/deny
         } elsif($lpriv =~ /^\d+$/) {  # else priv is all numeric
            $flags |= $lpriv & 0xfffc;  # mask off group/user and allow/deny
         } else {  # else priv name doesn't exist
            error_html::fatal("line $lineNum: I don't understand priv <tt>". cgi::htmlQuote($priv)."</tt>");
            exit;
         };
      }
      $acl .= pack("nn",$who,$flags);
   };
   return $acl;
}

#  Sadly, MySql before version 4 does not support multi-table updates,
#  thus we cannot set the ACL for branches with a WHERE condition referencing another table (dirent).
#  That turns out okay because our ACL design is too complicated for SQL anyway.
#  Although, it was designed that way in part because of the multi-table-update constraint.
#  Thus we must retrieve all the ACL and decide in perl which we can update, then we individually update them.
#
sub try_update_acl($$$) {
   my($file_id,$type,$parent) = (@_);
   return unless $cgi::form{'acl'};  # only if they hit a button
   session::require_login(); 
   my $formAcl = textToAcl($cgi::form{'acl'});
   my($where,$join) = ("","");
   if($cgi::form{'acl_all_files'} || $cgi::form{'acl_all_dirs'} || $cgi::form{'acl_recurse'}) {  # all files this directory (or recurse)
      if($type eq "dir") {
         $where = "parent = $file_id ";
      } else {
         $where = "parent = $parent ";
      }
      if($cgi::form{'acl_recurse'}) {  # all files this directory
         $where = "descendent.$where AND descendent.child = dirent.parent";
         $join = ",descendent ";
      } elsif($cgi::form{'acl_all_files'} && ! $cgi::form{'acl_all_dirs'}) {  # if only files and not dirs
         $where .= " AND dirent.type = 'file'";
      } elsif(! $cgi::form{'acl_all_files'} && $cgi::form{'acl_all_dirs'}) {  # if only dirs and not files
         $where .= " AND dirent.type = 'dir'";
      }
   } else { # else only one file
      $where = "dirent.file_id = $file_id ";
   };
   if($cgi::form{'branch'} ne "all") {  # if a selected branch
      $where .= "AND branchname = ". db::SqlQuote($cgi::form{'branch'});
   }
   # First do files (which have revisions)
   db::SendSQL("SELECT dirent.file_id, branchname, btime.acl, btime.owner ".
               "FROM dirent, btime $join".
               "WHERE dirent.file_id = btime.file_id AND $where AND btime.timestamp IS NULL");
   my(@file_ids,@branchnames);  # These ordered lists correspond.
   my $denied = 0;  #  Keep counts
   while(my($file_id,$branchname,$acl,$owner) = db::FetchSQLData()) {
      if(session::isPrivilegedTo($acl,0x100,$owner)) {
         push(@file_ids,$file_id);
         push(@branchnames,$branchname);
      } else {
         $denied++;
      }
   }
   # Next do directories (which don't have revisions)
   db::SendSQL("SELECT dirent.file_id, dirent.acl, dirent.owner ".
               "FROM dirent $join".
               "WHERE dirent.type = 'dir' AND $where ");
   my(@dir_ids);  
   while(my($dir_id,$acl,$owner) = db::FetchSQLData()) {
      if(session::isPrivilegedTo($acl,0x100,$owner)) {
         push(@dir_ids,$dir_id);
      } else {
         $denied++;
      }
   }
   if($#file_ids == -1 && $#dir_ids == -1) {
      error_html::fatal("Strangely, you didn't have privileges for any of the branches. Huh? ($denied denied)");
      exit;
   };
   my $update;
   if($cgi::form{'aclform'} =~ /top/i) {  # if we should append to the beginning
      $update = "acl = CONCAT(". db::SqlQuote($formAcl) . ",acl)";
   } elsif($cgi::form{'aclform'} =~ /bottom/i) { # if we should append to the end
      $update = "acl = CONCAT(acl,". db::SqlQuote($formAcl) . ")";
   } else { #if($cgi::form{'aclform'} =~ /set/i) {
      $update = "acl = ".db::SqlQuote($formAcl);
   }
   foreach my $index (0 .. $#file_ids) {
      db::SendSQL("UPDATE btime SET $update WHERE timestamp IS NULL AND file_id = ". $file_ids[$index] .
                  " AND branchname = ".db::SqlQuote($branchnames[$index]) );
   }
   foreach my $index (0 .. $#dir_ids) {
      db::SendSQL("UPDATE dirent SET $update WHERE file_id = ". $dir_ids[$index]);
   }
   return;
}

sub aclToText {
   my($acl) = @_;
   my $acltext = "";
   foreach my $rule (unpack("N*",$acl)) {
      my $who = $rule >> 16;
      my $flags = $rule & 0xffff;
      my $oldlen = length($acltext);
      if(($flags & 2)) {  # if user
         $acltext .= "user ";
         if($who == 0) {
            db::SendSQL("SELECT 'author'");
         } else {
            db::SendSQL("SELECT username FROM user WHERE userid = $who");
         }
      } else {  # else for group
         $acltext .= "group ";
         if($who == 0) {
            db::SendSQL("SELECT 'everyone'");
         } else {
            db::SendSQL("SELECT name FROM groups WHERE group_id = $who");
         }
      }
      my($whoname) = db::FetchSQLData();
      $who = $whoname if defined $whoname;
      $acltext .= "$who ";
      $acltext .= ($flags & 1) ? "allow " : "deny ";
      $acltext .= " " x (28-length($acltext)+$oldlen) if length($acltext)-$oldlen < 28;
      my $flagstext = "";
      foreach my $bit (sort {$b <=> $a} keys %privHelp) {  # for each priv sorted reverse numerically
         if($flags & $bit) {
            $flagstext .= "${$privHelp{$bit}}[0] ";
         } else {
            $flagstext .= " " x (length(${$privHelp{$bit}}[0]) + 1);
         }
      }
      $acltext .= $flagstext;
      $acltext .= "\n";
   }
   return $acltext;
}

sub access_html {
   my($filename,$file_id,$inforef) = @_; 

   try_update_acl($file_id,$inforef->{'type'},$inforef->{'parent'});
   common_html::content_type("text/html");
   my $branch = $cgi::form{'branch'} || "MAIN";
   if($inforef->{'type'} eq 'dir') {
      common_html::branchesWidget_keyword("dirent.parent = $file_id"); 
   } else {
      common_html::branchesWidget_keyword("dirent.file_id = $file_id"); 
   };
   $cgi::keyword{'loadfile'} = cgi::htmlQuote($filename);
   $cgi::keyword{'loadbranch'} = cgi::htmlQuote($branch);
   my @files = cvsdb::GetFiles("latest.acl, latest.owner","dirent.file_id = $file_id","fname",undef,$branch,$cgi::form{'label'});
   if(! defined $files[0]) {
      error_html::fatal("Internal error; File $filename not found");
      exit;
   }
   my($acl,$owner) = @{$files[0]};
   if($#files < 0) {  # if no files were found
      error_html::fatal("File $filename not found");
      exit;
   } elsif($#files > 0) {  # if multiple files were found
      error_html::fatal("Multiple files found?");
      exit;
   }
   session::require_privilege($acl,0x0200,$owner);  # read-p (read privileges)
   $acl = "" if ! defined $acl;
   $cgi::keyword{'debug'} = $config::debug ? "<br><font color=orange>(". unpack("H*",$acl) .")</font><br>\n" :"";
   $cgi::keyword{'acl'} = cgi::htmlQuoteForPre(aclToText($acl));
   $cgi::keyword{'aclPrefix'} = cgi::htmlQuoteForPre(aclToText($config::aclPrefix));
   $cgi::keyword{'aclSuffix'} = cgi::htmlQuoteForPre(aclToText($config::aclSuffix));
   $cgi::keyword{'loglink'} = cgi::Link("$filename---log.html","after=".::time2str("%b %e",$db::now - 7*24*3600));

   # Do the help for attributes
   $cgi::keyword{'help'} = "<table>\n<tr bgcolor=cccccc><td><font color=224477>code</font></td>\n".
                         "  <td><font color=224477>priv</font></td><td><font color=224477>description</font></td></tr>\n";
   my $rowIndex = 0;
   foreach my $helpkey (sort {$b <=> $a} keys %privHelp) {  # for each privilege sorted reverse numerically
      my $color = ($rowIndex++ & 1) ? "ffffff" : "dddddd";  # alternating rows
      my($attr,$help) = @{$privHelp{$helpkey}};
      my $hex = sprintf("0x%04x",$helpkey);
      $cgi::keyword{'help'} .= "<tr bgcolor=$color><td>$hex</td><td><tt>$attr</tt></td><td>$help</td></tr>\n";
   };
   $cgi::keyword{'help'} .= "</table>\n";

   $cgi::keyword{'formurl'} = cgi::Uri("$filename---access.html");
   $cgi::keyword{'acl_comments'} = $cgi::form{'acl_comments'} || "";
   $cgi::keyword{'acl_new_files'} = $cgi::form{'acl_new_files'} || "";
   $cgi::keyword{'acl_all_files'} = $cgi::form{'acl_all_files'} || "";
   $cgi::keyword{'acl_all_dirs'} = $cgi::form{'acl_all_dirs'} || "";
   $cgi::keyword{'acl_recurse'} = $cgi::form{'acl_recurse'} || "";
   print cgi::expand_template_keywords("template/access-$inforef->{'type'}-template.html");
}

1;

