#!/usr/bin/perl -w
# -*- Mode: perl; indent-tabs-mode: nil -*-
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is mozilla.org code.
#
# The Initial Developer of the Original Code is Holger
# Schurig. Portions created by Holger Schurig are
# Copyright (C) 1999 Holger Schurig. All
# Rights Reserved.
#
# Contributor(s): Holger Schurig <holgerschurig@nikocity.de>
#                 Terry Weissman <terry@mozilla.org>
#                 Dan Mosedale <dmose@mozilla.org>
#                 Dave Miller <justdave@syndicomm.com>
#                 Frederick Dean <software@fdd.com>
#
#
#
#
# Hey, what's this?
#
# 'checksetup.pl' is a script that is supposed to run during installation
# time and also after every upgrade.
#
# The goal of this script is to make the installation even more easy.
# It does so by doing things for you as well as testing for problems
# early.
#
# And you can re-run it whenever you want. Especially after DCVS
# gets updated you SHOULD rerun it. Because then it may update your
# SQL table definitions so that they are again in sync with the code.
#
# So, currently this module does:
#
#     - check for required perl modules
#     - set defaults for local configuration variables
#     - create and populate the data directory after installation
#     - set the proper rights for the *.cgi, *.html ... etc files
#     - check if the code can access MySQL
#     - creates the database 'bugs' if the database does not exist
#     - creates the tables inside the database if they don't exist
#     - automatically changes the table definitions of older BugZilla
#       installations
#     - populates the groups
#     - put the first user into all groups so that the system can
#       be administrated
#     - changes already existing SQL tables if you change your local
#       settings, e.g. when you add a new platform
#
# People that install this module locally are not supposed to modify
# this script. This is done by shifting the user settable stuff into
# a local configuration file 'config.pm'. When this file get's
# changed and 'checkconfig.pl' will be re-run, then the user changes
# will be reflected back into the database.
#
# Developers however have to modify this file at various places. To
# make this easier, I have added some special comments that one can
# search for.
#
#     To                                               Search for
#
#     add/delete local configuration variables         --LOCAL--
#     check for more prerequired modules               --MODULES--
#     change the defaults for local configuration vars --LOCAL--
#     update the assigned file permissions             --CHMOD--
#     add more MySQL-related checks                    --MYSQL--
#     change table definitions                         --TABLE--
#     add more groups                                  --GROUPS--
#     create initial administrator account            --ADMIN--
#
# Note: sometimes those special comments occur more then once. For
# example, --LOCAL-- is at least 3 times in this code!  --TABLE--
# also is used more than once. So search for every occurence!
#


use diagnostics;
use strict;

###########################################################################
# Check required module
###########################################################################

#
# Here we check for --MODULES--
#

print "\nChecking perl modules ...\n";
unless (eval "require 5.004") {
    die "Sorry, you need at least Perl 5.004\n";
}

# vers_cmp is adapted from Sort::Versions 1.3 1996/07/11 13:37:00 kjahds,
# which is not included with Perl by default, hence the need to copy it here.
# Seems silly to require it when this is the only place we need it...
sub vers_cmp {
  if (@_ < 2) { die "not enough parameters for vers_cmp" }
  if (@_ > 2) { die "too many parameters for vers_cmp" }
  my ($a, $b) = @_;
  my (@A) = ($a =~ /(\.|\d+|[^\.\d]+)/g);
  my (@B) = ($b =~ /(\.|\d+|[^\.\d]+)/g);
  my ($A,$B);
  while (@A and @B) {
    $A = shift @A;
    $B = shift @B;
    if ($A eq "." and $B eq ".") {
      next;
    } elsif ( $A eq "." ) {
      return -1;
    } elsif ( $B eq "." ) {
      return 1;
    } elsif ($A =~ /^\d+$/ and $B =~ /^\d+$/) {
      return $A <=> $B if $A <=> $B;
    } else {
      $A = uc $A;
      $B = uc $B;
      return $A cmp $B if $A cmp $B;
    }
  }
  @A <=> @B;
}

# This was originally clipped from the libnet Makefile.PL, adapted here to
# use the above vers_cmp routine for accurate version checking.
sub have_vers {
  my ($pkg, $wanted) = @_;
  my ($msg, $vnum, $vstr);
  no strict 'refs';
  printf("Checking for %15s %-9s ", $pkg, !$wanted?'(any)':"(v$wanted)");

  eval { my $p; ($p = $pkg . ".pm") =~ s!::!/!g; require $p; };

  $vnum = ${"${pkg}::VERSION"} || ${"${pkg}::Version"} || 0;
  $vnum = -1 if $@;

  if ($vnum < 0) {
    $vstr = "not found";
  }
  elsif ($vnum > 0) {
    $vstr = "found v$vnum";
  }
  else {
    $vstr = "found unknown version";
  }

  my $vok = (vers_cmp($vnum,$wanted) > -1);
  print ((($vok) ? "ok: " : " "), "$vstr\n");
  return $vok;
}

# Check versions of dependencies.  0 for version = any version acceptible

my @missing = ();
unless (have_vers("DBI","1.13"))          { push @missing,"DBI" }
unless (have_vers("DBD::mysql","1.2209")) { push @missing,"DBD::mysql" }
unless (have_vers("Date::Parse",0))       { push @missing,"Date::Parse" }
unless (have_vers("Date::Format",0))       { push @missing,"Date::Format" }
unless (have_vers("Digest::MD5",0))       { push @missing,"Digest::MD5" }

# If CGI::Carp was loaded successfully for version checking, it changes the
# die and warn handlers, we don't want them changed, so we need to stash the
# original ones and set them back afterwards -- justdave@syndicomm.com
my $saved_die_handler = $::SIG{__DIE__};
my $saved_warn_handler = $::SIG{__WARN__};
unless (have_vers("CGI::Carp",0))    { push @missing,"CGI::Carp" }
$::SIG{__DIE__} = $saved_die_handler;
$::SIG{__WARN__} = $saved_warn_handler;

if (@missing > 0) {
    print "\n\n";
    print "You are missing some Perl modules which are required by DCVS.\n";
    print "They can be installed by running (as root) the following:\n";
    foreach my $module (@missing) {
        print "   perl -MCPAN -e 'install \"$module\"'\n";
    }
    print "\n";
    exit;
}

###########################################################################
# Check and update local configuration
###########################################################################

#
# This is quite tricky. But fun!
#
# First we read the file 'config.pm'. Then we check if the variables we
# need are defined. If not, config.pm will be amended by the new settings
# and the user informed to check this. The program then stops.
#
# Why do it this way around?
#
# Assume we will enhance DCVS and eventually more local configuration
# stuff arises on the horizon.
#
# But the file 'config.pm' is not in the DCVS archive or tarfile. You
# know, we never want to overwrite your own version of 'config.pm', so
# we can't put it into the archive/tarfile, can we?
#
# Now, we need a new variable. We simply add the necessary stuff to checksetup.
# The user get's the new version of DCVS from the archive, runs checksetup
# and checksetup finds out "Oh, there is something new". Then it adds some
# default value to the user's local setup and informs the user to check that
# to see if that is what the user wants.
#
# Cute, ey?
#

print "\nChecking user setup ...\n";
do 'config.pm';
my $newstuff = "";
sub LocalVar ($$)
{
    my ($name, $definition) = @_;
    return if ($config::{$name}); # if localconfig declared it, we're done.
    $newstuff .= "\n   \$config::" . $name;
    if(! -e "config.pm") {
       open FILE, '>config.pm' or die "\n\tI cannot open config.pm\n\n";
       print(FILE "#!/usr/bin/perl -w -T\n".
                  "use strict;\n".
                  "package config;\n".
                  "\n".
                  "# Don't check this file into CVS.\n".
                  "# It is generated (grown) by checkconfig.pl.\n".
                  "\n");
    }
    open FILE, '>>config.pm' or die "\n\tI cannot open config.pm\n\n";
    print FILE $definition;
    close FILE;
}

#
# Set up the defaults for the --LOCAL-- variables below:
#
    
#LocalVar('webservergroup', '
 #
 # This is the group your web server runs on.
 # If you have a windows box, ignore this setting.
 # If you do not wish for checksetup to adjust the permissions of anything,
 # set this to "".
 # If you set this to anything besides "", you will need to run checksetup.pl
 # as root.
#our $webservergroup = "www";
#');

LocalVar('db_host', '
#
# How to access the SQL database:
our $db_host = "localhost";         # where is the database?
our $db_port = 3306;                # which port to use
our $db_name = "dcvs";              # name of the MySQL database
our $db_user = "dcvs";              # user to attach to the MySQL database
our $db_pass = "";                   # CHANGE THIS PASSWORD!
');

LocalVar('db_check', '
#
# Should checksetup.pl try to check if your MySQL setup is correct?
# (with some combinations of MySQL/Msql-mysql/Perl/moonphase this doesn\'t work)
our $db_check = 1;
');

LocalVar('timezone', '
#
# What timezone should the web pages report in...
our $timezone = "CST";
');

LocalVar('timeformat', '
#
# What timeformat should the web pages report with?
# Run "perldoc Date::Format" for the meaning of escape codes.
our $timeformat = "%a %b %d, %Y %H:%M:%S %Z";
');

LocalVar('root', '
#
# What (virtual) root should we pretend to use for pserver.
# The clients will use this when setting their CVSROOT environment variable.
# Many version of the cvs client have troubles updating if this is "/".
# It should begin with a slash but not end with one.
our $root = "/cvsroot";
');

LocalVar('debug', '
#
# This should not be enabled on a live system.
our $debug = 0;
');

LocalVar('maxFileSize', '
#
# We will not accept files larger than this.
our $maxFileSize = 1000000;
');

LocalVar('bugzillaLink', '
#
# URL to convert commit comment numbers to links with.
# The number is appended to this should be a GET form with the bug_id parameter last.
# If nonempty, the links will not be created.
our $bugzillaLink = "http:://bugzilla/show_bug.cgi?id=";
');

LocalVar('minNumBugDigits', '
#
#  If you are willing to sacrifice the first 100 bugs... :-)
our $minNumBugDigits = 3;
');

LocalVar('aclPrefix', '
#
#  This is prepended to the beginning of every ACL list. 
#  The default is 0x0001fffd (group administrators allow all)
our $aclPrefix = pack("H*","0001fffd");
');

LocalVar('aclSuffix', '
#
#  This is appended to the end of every ACL list.
#  The default is 0x0001fffd (group everyone allow move-l create-l create-b append-c read-c commit read-f visible)
#                 0x0001fffc (group everyone deny all)
our $aclSuffix = pack("H*","000054bd".""."0000fffc");
');

LocalVar('scriptUser', '
#
#  The config.pm contains the database password for dcvs, so we need to protect it.
#  Anyone with that password roots version control.
#  Our plan makes config.pm only readable by $scriptUser, and makes
#  dcvs.cgi SUID $scriptUser.  
#
#  Better strategies exist where the scripts run without privileges to write
#  to the scripts (i.e. also not ownership).
#  Instead, they use group permission to read the password.  
#  A chroot jail would also help (athough lookout for load and save of RCS files).
#  If you do not want checksetup.pl to do file permissions, set $scriptUser to undef.
#
#  WARNING: If you use the userid of your web server, anyone who can write a
#  cgi (or exploit a cgi hole) can view the database password.
our $scriptUser = "dcvs";
');

LocalVar('minUseridToInheritFromOS', '
#
#  When authenticating a user, and the username is not already in
#  the database, create a user account if the OS has a user by the
#  same name and the numeric userid is greater than $minUseridToInheritFromOS.
#  These inherited accounts will always go to the OS to check the password each time.
#  WARNING: we do not check the shell of the OS user account.
#  For this feature your SUID perl must work for pwverify.pl.
#  A setting of undef means "inherit no accounts"
our $minUseridToInheritFromOS = 500;
');

LocalVar('emailDomainNameForInherited', '
#
#  Email domain name for OS inherited accounts.
our $emailDomainNameForInherited = \'@localhost\';
');

if ($newstuff ne "") {
    print "\n".
          "This version of dcvs contains some variables that you may\n".
          "want to adapt to your local settings...\n".
          "$newstuff\n".
          "\n".
          "Please edit the file 'config.pm' and rerun checksetup.pl\n\n";
    exit;
}

# 2000-Dec-18 - justdave@syndicomm.com - see Bug 52921
# This is a hack to read in the values defined in config.pm without getting
# them predeclared at compile time if they're missing from config.pm.
# Ideas swiped from pp. 281-282, O'Reilly's "Programming Perl 2nd Edition"
# Note that we won't need to do this in globals.pl because globals.pl couldn't
# care less whether they were defined ahead of time or not. 
my $my_db_check = ${*{$config::{'db_check'}}{SCALAR}};
my $my_db_host = ${*{$config::{'db_host'}}{SCALAR}};
my $my_db_port = ${*{$config::{'db_port'}}{SCALAR}};
my $my_db_name = ${*{$config::{'db_name'}}{SCALAR}};
my $my_db_user = ${*{$config::{'db_user'}}{SCALAR}};
my $my_db_pass = ${*{$config::{'db_pass'}}{SCALAR}};
#my $my_webservergroup = ${*{$config::{'webservergroup'}}{SCALAR}};
my $my_scriptUser = ${*{$config::{'scriptUser'}}{SCALAR}};
my $my_minuser = ${*{$config::{'minUseridToInheritFromOS'}}{SCALAR}};

###########################################################################
# Update the @INC path in the scripts
###########################################################################

use Cwd;

$0 =~ /(.*)\/[^\/]+$/;  # match path before executable
my $libPath = $1;
if($libPath !~ /^\//) { # if path doesn't begin with slash
   $libPath = Cwd::abs_path($libPath);
}

print("\n");
my @scripts = qw( dcvs.cgi dcvsd.pl sanity.pl loadrcs.pl savercs.pl rm-dcvs.pl );
push @scripts, glob("$libPath/test/*.pl");  # do perl scripts in test directory
for my $file (@scripts) {
   $file = "$libPath/$file" unless $file =~ /^\//;  # unless begins with slash
   open(INPL,"<$file") or die;
   my $numChars = read(INPL,my $filedata,10000); 
   if($filedata =~ s/^(\s*BEGIN\s*\{\s*(?:push|unshift)[\(\s]+\@INC\s*,\s*")(.*)"/$1$libPath"/m) {
      next if($2 eq $libPath);  # if old value is same as current
   } else {
      die("\nI couldn't find a line like ...\n".
             "   BEGIN { push \@INC, \"/path/to/libs\" };\n".
             "...in file $file to update\n");
   };
   print("Updating \@INC path of $file\n");
   open(OUTPL,">$file.$$") or die;
   while(1) {
      defined($numChars) or die;
      last if $numChars == 0;
      print(OUTPL $filedata); 
      $numChars = read(INPL,$filedata,10000); 
   }
   close(INPL) or die;
   close(OUTPL) or die;
   rename("$file.$$","$file") or die; 
}
close(INPL);
close(OUTPL);

###########################################################################
# File permissions
###########################################################################

sub fixOwnership {
   my($filename,$username) = (@_);

   my $uid = $username;
   if($uid !~ /^\d+$/) {  # if my_scriptUser is not all numeric
      $uid   = getpwnam($username);
      if(! defined $uid) {
         die("Operating system (i.e. not database) user $username does not exist.");
      };
   };
   my $group = (stat($filename))[5];  # get group so we can leave it unchanged
   chown($uid,$group,$filename);  # okay if fails
   my $file_uid = (stat($filename))[4];   
   if($uid != $file_uid) {
      print(STDERR "The file $filename is not owned by $username ($uid != $file_uid)\n".
                   "and I can't fix it, but you can with...\n".
                   "\tsu -c \"chown $username $filename\"\n".
                   "...or rerunning checksetup.pl as root.\n");
      exit -1;
   }
}

sub fixMode {
   my($mode,@filenames) = (@_);

   foreach my $filename (@filenames) {
      chmod $mode, $filename;
      my $actualMode = (stat($filename))[2] & 07777;
      if($actualMode != $mode) {
         printf(STDERR "The file $filename has the wrong permissions, and \n".
                       "I can't fix it, but you can with...\n".
                       "\tsu -c \"chmod %o $filename\"\n".
                       "...or rerunning checksetup.pl as root.\n",$mode);
         exit -1;
      }
   };
}

if(defined $my_scriptUser) {
   fixOwnership("config.pm",$my_scriptUser);
   fixMode( 0600, "config.pm"); # -rw-------
   fixOwnership("dcvs.cgi",$my_scriptUser);
   fixMode( 0755, @scripts);    # -rwxr-xr-x
   fixMode(04755, "dcvs.cgi");  # -rwsr-xr-x
   fixMode( 0755, ".");         # -rwxr-xr-x
}

if(defined $my_minuser) {  # if we are allowing OS user acounts for DCVS
   fixOwnership("pwverify.pl","root");
   fixMode( 04755, "pwverify.pl"); # -rwsr-xr-x
}

###########################################################################
# Check MySQL setup
###########################################################################

my $bigHelp = <<"EOF"
This might have several reasons:

* MySQL is not running.
* MySQL is running, but the rights are incorrect. 
* You goofed the parameters in config.pm.
* There is an subtle problem with Perl, DBI, DBD::mysql and MySQL. 
   If all else fails, set '\$db_check' in config.pm to zero.

If you have not already, the rights can be set up with a statement like...

mysql -p --user=root --host=127.0.0.1 -e "GRANT SELECT,INSERT,UPDATE,DELETE,
INDEX,ALTER,CREATE,DROP,REFERENCES ON dcvs.* TO dcvs\@127.0.0.1 IDENTIFIED BY
'dcvs_db_password'; FLUSH PRIVILEGES;"

... but choose a better password.  The passord is asks for is not the
one being set, but the root of the database (different than root of unix).
This too must not be left blank.

EOF
;

#
# Check if we have access to --MYSQL--
#

# This settings are not yet changeable, because other code depends on
# the fact that we use MySQL and not, say, PostgreSQL.

my $db_base = 'mysql';

# No need to "use" this here.  It should already be loaded from the
# version-checking routines above, and this file won't even compile if
# DBI isn't installed so the user gets nasty errors instead of our
# pretty one saying they need to install it. -- justdave@syndicomm.com
#use DBI;

# get a handle to the low-level DBD driver
my $drh = DBI->install_driver($db_base)
    or die "Can't connect to the $db_base. Is the database installed and up and running?\n";

if ($my_db_check) {
    # Do we have the database itself?

    my $sql_want = "3.22.5";  # minimum version of MySQL

# original DSN line was:
#    my $dsn = "DBI:$db_base:$my_db_name;$my_db_host;$my_db_port";
# removed the $db_name because we don't know it exists yet, and this will fail
# if we request it here and it doesn't. - justdave@syndicomm.com 2000/09/16
    my $dsn = "DBI:$db_base:;$my_db_host;$my_db_port";
    my $dbh = DBI->connect($dsn, $my_db_user, $my_db_pass)
      or die "\nI can't connect to the $db_base database. $bigHelp";
    printf("Checking for %15s %-9s ", "MySQL Server", "(v$sql_want)");
    my $qh = $dbh->prepare("SELECT VERSION()");
    $qh->execute;
    my ($sql_vers) = $qh->fetchrow_array;
    $qh->finish;

    # Check what version of MySQL is installed and let the user know
    # if the version is too old to be used with DCVS.
    if ( vers_cmp($sql_vers,$sql_want) > -1 ) {
        print "ok: found v$sql_vers\n\n";
    } else {
        die "Your MySQL server v$sql_vers is too old./n" . 
            "   DCVS requires version $sql_want or later of MySQL.\n" . 
            "   Please visit http://www.mysql.com/ and download a newer version.\n";
    }

    my @databases = $dbh->func('_ListDBs');
    unless (grep /^$my_db_name$/, @databases) {
       print "Creating database $my_db_name ...\n";
       $drh->func('createdb', $my_db_name, "$my_db_host:$my_db_port", $my_db_user, $my_db_pass, 'admin')
            or die " The '$my_db_name' database is not accessible. $bigHelp";
    }
    $dbh->disconnect if $dbh;
}

# now get a handle to the database:
my $connectstring = "dbi:$db_base:$my_db_name:host=$my_db_host:port=$my_db_port";
my $dbh = DBI->connect($connectstring, $my_db_user, $my_db_pass)
    or die "Can't connect to the table '$connectstring'.\n",
           "Have you read the DCVS Guide in the doc directory?  Have you read the doc of '$db_base'?\n";

END { $dbh->disconnect if $dbh }





###########################################################################
# Table definitions
###########################################################################

#
# The following hash stores all --TABLE-- definitions. This will be used
# to automatically create those tables that don't exist. 
#
# If you want to intentionally recreate the table, you can drop a table and re-run
# checksetup, e.g. like this:
#
#    $ mysql dcvs
#    mysql> drop table votes;
#    mysql> exit;
#    $ ./checksetup.pl
#
# If you change one of those field definitions, then also go below to the
# next occurence of the string --TABLE-- (near the end of this file) to
# add the code that updates older installations automatically.
#


my %table;
 
$table{session} =
   'cookie varchar(22) primary key not null,
    userid mediumint unsigned not null,
    lastused int unsigned not null,
    remoteip int unsigned not null,
    isvalid mediumint unsigned not null default 0,

    index(lastused)';
 
# Special group_id 
#    0   anonymous
# Groups is a packed list of groups, each one 16-bits long, network byte order. (group 0 not needed)
$table{user} =
   'userid mediumint unsigned not null auto_increment primary key,
    email varchar(255) not null,
    username varchar(32) not null,
    cryptpassword varchar(47),
    realname varchar(255),
    disabledtext mediumtext not null,
    groups mediumblob not null,
 
    index(userid),
    unique(username)';

# Special group_id 
#     0     everyone
#     1     administrators
#    >100   user created groups
# Table name "group" singular is not allowed (reserved word).
$table{groups} =
   'group_id mediumint unsigned not null auto_increment primary key,
    name mediumtext not null,
    descr mediumtext not null,

    index(group_id)';


# "directory entry"  (i.e. files, directories, symlink, etc)
# "parent" is a file_id for another file.  0 for at root.
# "head" is a revision num
# "path" begins with a / but has no trailing one
# "path" is only populated for directories.  It includes fname.
# "file_id" should really be "dirent_id"
# "acl" is only valid if the file_id has no revisions (i.e. not version controlled).
# file_id=0 is special. It is the root directory.
$table{dirent} =
   'file_id mediumint unsigned not null auto_increment primary key, 
    fname varchar(128) not null,
    parent mediumint not null,
    type enum("file","dir","link","dev"),
    comment mediumtext,
    head mediumtext,    # is a revision num not rev_id
    path mediumtext,  # only populated for directories.  It includes fname
    expand mediumtext,  # not yet used, since we expand nothing yet 
    mode mediumtext,
    acl mediumblob not null,    # only valid when no revisions (and thus btime records) exist
    owner mediumint unsigned,   # only valid when no revisions (and thus btime records) exist

    unique(parent,fname)';

# The root directory (parent=0) doesn't get entries.
# We compare descendent.child to dirent.parent so we do NOT need entries for specific files.
# MySQL doesn't handle 'OR' clauses so well, so every directory also lists itself as its own child.
$table{descendent} =
   'parent mediumint unsigned not null,
    child mediumint unsigned not null,
    
    unique(parent,child)';

#  For each file_id, the timestamp and rev_id MUST be in the revision num order.
#  Very important: the rev_id must be in order.
#  "branches" are not used (is a figment of rcs files).
#  "branchnum" is empty for the trunk, otherwise it is revision with last number and dot removed.
#  RCS files store this odd number of nubers like X.Y.Z as X.Y.0.Z, go figure.
$table{revision} =
   'rev_id mediumint unsigned not null auto_increment primary key, 
    file_id mediumint not null,
    revision varchar(64) not null,
    comment_id mediumint not null,
    state mediumtext,
    branches mediumtext,
    branchnum varchar(64),
    next mediumtext,
    timestamp int unsigned not null,
    text mediumblob,
    md5 varchar(22),
    length int unsigned not null,
    added mediumint,
    removed mediumint,

    index(file_id,rev_id),
    index(file_id,branchnum,rev_id),
    unique(file_id,revision)';


$table{comment} =
   'comment_id mediumint unsigned not null auto_increment primary key, 
    descr mediumtext,
    timestamp int unsigned not null,
    author varchar(32) not null,  # matches username not userid of user table
    acl mediumblob not null
    ';

$table{label} =
   'file_id mediumint not null,
    label varchar(64) not null,
    revision varchar(64) not null,

    unique(file_id,label)';

# For the following table only applies before timestamp. (null means no time limit)
# Generally, every branch has one and only one NULL timestamp (per file)
$table{btime} =
   'file_id mediumint not null,
    branchname varchar(64) not null,
    branchnum varchar(64) not null,
    timestamp int unsigned,
    acl mediumblob not null,   # only valid if timestamp is NULL
    owner mediumint unsigned,  # only valid if timestamp is NULL

    index(file_id,timestamp,branchname),
    index(file_id,branchname,timestamp),
    unique(file_id,branchnum,branchname)';

=pod unused
#  THE FOLLOWING TABLE IS STILL UNUSED
#
#  For performance we will lock on directories, not files.
#  Locking one directory locks all of its children.
#  Locks will only be for a selected branch.
#  Once we have composed every diff, we will mysql-lock the tables and apply them.
#  This table is for when we start composing the diffs and when lock them.
#  This table only needs to be checked by committers.
#
$table{commitlock} =
   'lock_id mediumint unsigned not null auto_increment primary key, 
    parent mediumint not null,  
    author varchar(64) not null,
    branchname varchar(64) not null,
    pid varchar(64) not null,
    timestamp int unsigned not null,

    index(parent) ';
=cut

###########################################################################
# Create tables
###########################################################################

# Get a list of the existing tables (if any) in the database
my @tables = map { $_ =~ s/.*\.//; $_ } $dbh->tables;
#print 'Tables: ', join " ", @tables, "\n";

# go throught our %table hash and create missing tables
while (my ($tabname, $fielddef) = each %table) {
    next if grep /^$tabname$/, @tables;
    print "Creating table $tabname ...\n";

    $dbh->do("CREATE TABLE $tabname (\n$fielddef\n)")
        or die "Could not create table '$tabname'. Please check your '$db_base' access.\n";
}

###########################################################################
# Populate dirent table
###########################################################################

sub AddDir {
   my($fname,$path,$parent,$file_id,$acl) = @_;

   my $sth = $dbh->prepare("SELECT file_id FROM dirent WHERE file_id=$file_id");
   $sth->execute;
   return if ($sth->rows);  # if it already exists, do nothing more (sanity.pl will fix it if need be)
   $sth = $dbh->prepare("DELETE FROM dirent WHERE fname = ? AND parent = ?");
   $sth->execute($fname,$parent);  # avoid complaints about duplicates
   $sth = $dbh->prepare("INSERT dirent SET fname = ?, path = ?, parent = ?, type = 'dir', acl = ?");
   $sth->execute($fname,$path,$parent,$acl);
   if(defined $file_id) { 
      $sth = $dbh->prepare("SELECT LAST_INSERT_ID()");
      $sth->execute();  
      my($last_id) = $sth->fetchrow_array();
      if($last_id != $file_id) {  # if the file_id is wrong
         $sth = $dbh->prepare("UPDATE dirent SET file_id = $file_id WHERE file_id = $last_id");
         $sth->execute();  # fix the file_id
      };
   };
}

AddDir("","/",-1,0,pack("H*","0002fffd"));   # The ACL edit stuff requires this one to exist
#   ACL  0x0002fffd is (user owner allow move-l create-l create-b append-c read-c commit read-f visible)

#AddDir("CVSROOT","/CVSROOT",0,undef);

###########################################################################
# Populate groups table
###########################################################################

sub GroupExists ($)
{
    my ($name) = @_;
    my $sth = $dbh->prepare("SELECT name FROM groups WHERE name='$name'");
    $sth->execute;
    if ($sth->rows) {
        return 1;
    }
    return 0;
}

#
# This subroutine checks if a group exist. If not, it will be automatically
# created with the next available bit set
sub AddGroup {
    my ($id, $name, $desc) = @_;

    return if GroupExists($name);
    print "Adding group $name ...\n";
    my $sth1 = $dbh->prepare('DELETE FROM groups WHERE group_id = ?');
    $sth1->execute($id);
    my $sth2 = $dbh->prepare('INSERT INTO groups (group_id, name, descr) VALUES (?, ?, ?)');
    $sth2->execute($id, $name, $desc);
    my $sth3 = $dbh->prepare('UPDATE groups SET group_id = ? WHERE name = ?');
    $sth3->execute($id, $name);
}


#
# dcvs uses --GROUPS-- to assign various rights to its users. 
AddGroup(0,'everyone','everyone');
AddGroup(1,'administrators','administrators who can edit users, etc.');


###########################################################################
# Detect changed local settings
###########################################################################

sub GetFieldDef ($$)
{
    my ($table, $field) = @_;
    my $sth = $dbh->prepare("SHOW COLUMNS FROM $table");
    $sth->execute;

    while (my $ref = $sth->fetchrow_arrayref) {
        next if $$ref[0] ne $field;
        return $ref;
   }
}

sub GetIndexDef ($$)
{
    my ($table, $field) = @_;
    my $sth = $dbh->prepare("SHOW INDEX FROM $table");
    $sth->execute;

    while (my $ref = $sth->fetchrow_arrayref) {
        next if $$ref[2] ne $field;
        return $ref;
   }
}

sub CountIndexes ($)
{
    my ($table) = @_;
    
    my $sth = $dbh->prepare("SHOW INDEX FROM $table");
    $sth->execute;

    if ( $sth->rows == -1 ) {
      die ("Unexpected response while counting indexes in $table:" .
           " \$sth->rows == -1");
    }
    
    return ($sth->rows);
}

sub DropIndexes ($)
{
    my ($table) = @_;
    my %SEEN;

    # get the list of indexes
    #
    my $sth = $dbh->prepare("SHOW INDEX FROM $table");
    $sth->execute;

    # drop each index
    #
    while ( my $ref = $sth->fetchrow_arrayref) {
      
      # note that some indexes are described by multiple rows in the
      # index table, so we may have already dropped the index described
      # in the current row.
      # 
      next if exists $SEEN{$$ref[2]};

      my $dropSth = $dbh->prepare("ALTER TABLE $table DROP INDEX $$ref[2]");
      $dropSth->execute;
      $dropSth->finish;
      $SEEN{$$ref[2]} = 1;

    }

}
#
# Check if the enums in the bugs table return the same values that are defined
# in the various locally changeable variables. If this is true, then alter the
# table definition.
#

sub CheckEnumField ($$@)
{
    my ($table, $field, @against) = @_;

    my $ref = GetFieldDef($table, $field);
    #print "0: $$ref[0]   1: $$ref[1]   2: $$ref[2]   3: $$ref[3]  4: $$ref[4]\n";
    
    $_ = "enum('" . join("','", @against) . "')";
    if ($$ref[1] ne $_) {
        print "Updating field $field in table $table ...\n";
        $_ .= " NOT NULL" if $$ref[3];
        $dbh->do("ALTER TABLE $table
                  CHANGE $field
                  $field $_");
    }
}

###########################################################################
# Create Administrator  --ADMIN--
###########################################################################

sub bailout {   # this is just in case we get interrupted while getting passwd
    system("stty","echo"); # re-enable input echoing
    exit 1;
}
sub getPassword {
    $SIG{HUP}  = \&bailout;
    $SIG{INT}  = \&bailout;
    $SIG{QUIT} = \&bailout;
    $SIG{TERM} = \&bailout;
    system("stty","-echo");  # disable input echoing
    my $pass = <STDIN>;
    system("stty","echo"); # re-enable input echoing
    $SIG{HUP}  = 'DEFAULT'; # and remove our interrupt hooks
    $SIG{INT}  = 'DEFAULT';
    $SIG{QUIT} = 'DEFAULT';
    $SIG{TERM} = 'DEFAULT';
    print("\n");
    return $pass;
}

sub adminList {
   my @admins = ();
   my $sth = $dbh->prepare("SELECT username, groups FROM user");
   $sth->execute;
   while( my($username,$groups) = $sth->fetchrow_array ) {
      foreach my $group (unpack("n*",$groups)) {
         if($group == 1) {  # group 1 is the administrators group 
            push(@admins,$username);
            last;  # stop doing groups for this user, and do next user
         };
      };
   };
   return @admins;
}

my @admins = adminList();
if($#admins > -1) {
   print("\nYou already have some administrators... ". join(" ",@admins) ."\n");
} else {  # else we don't have an administrator but need one.
   print("\nYou do not have an administrator account, so we need to set one up.\n".
         "If you make a mistake, use control-C and run checksetup.pl again.\n".
         "What username will the administrator use? ");
   my $username = <STDIN>;
   chomp($username);
   my $sth = $dbh->prepare("SELECT userid, realname, groups FROM user WHERE username = ?");
   $sth->execute($username);
   if( my($userid,$realname,$groups) = $sth->fetchrow_array ) {
      print("\nUser $username ($realname) already exists.\nShould we promote her to administrator? ");
      my $confirm = <>;
      exec($0) unless $confirm =~ /^y/i;  # restart unless response begins with y
      $groups .= pack("n",1);
      my $sth = $dbh->prepare("UPDATE user SET groups = ? WHERE userid = ?");
      $sth->execute($groups,$userid);
   } else {  # else the user does not exist
      print("\nWhat is ${username}'s real name? ");
      my $realname = <STDIN>;
      chomp($realname);
      print("\nWhat is ${username}'s email address? ");
      my $email = <STDIN>;
      chomp($email);
      my($pass) = ("a");
      while(1) {  # while we try to agree on a password
         print("\nWhat is ${username}'s password? ");
         $pass = getPassword();
         if(length($pass) < 8) {
            print("\nYour password was too short. Please try again with 8 or more chars.\n");
            next;
         }
         if($pass =~ /^[a-zA-Z0-9]+$/) {  # is all alphanumeric
            print("\nYour password was all alphanumeric. Please try again.\n");
            next;
         }
         print("again? ");
         my $pass2 = getPassword();
         if($pass ne $pass2) {
            print("\nThose passwords didn't match, please try again.\n");
            next;
         };
         last;   # stop this password madness
      };
      require useradmin;  # can't be use, we need this at run (not compile) time.
      useradmin::set_user($username,$pass,$realname,$email,pack("n",1));
      #useradmin::set_user($username,$pass,$realname,$email,useradmin::groupList("administrators"));
      print("Okay, I have created an administrator account for $username\n\n");
   };
}

###########################################################################
# Update the tables to the current definition
###########################################################################

#
# As time passes, fields in tables get deleted, added, changed and so on.
# So we need some helper subroutines to make this possible:
#

sub ChangeFieldType ($$$)
{
    my ($table, $field, $newtype) = @_;

    my $ref = GetFieldDef($table, $field);
    #print "0: $$ref[0]   1: $$ref[1]   2: $$ref[2]   3: $$ref[3]  4: $$ref[4]\n";

    my $oldtype = $ref->[1];
    if (! $ref->[2]) {
        $oldtype .= qq{ not null};
    }
    if ($ref->[4]) {
        $oldtype .= qq{ default "$ref->[4]"};
    }

    if ($oldtype ne $newtype) {
        print "Updating field type $field in table $table ...\n";
        print "old: $oldtype\n";
        print "new: $newtype\n";
#        'not null' should be passed as part of the call to ChangeFieldType()
#        $newtype .= " NOT NULL" if $$ref[3];
        $dbh->do("ALTER TABLE $table
                  CHANGE $field
                  $field $newtype");
    }
}

sub RenameField ($$$)
{
    my ($table, $field, $newname) = @_;

    my $ref = GetFieldDef($table, $field);
    return unless $ref; # already fixed?
    #print "0: $$ref[0]   1: $$ref[1]   2: $$ref[2]   3: $$ref[3]  4: $$ref[4]\n";

    if ($$ref[1] ne $newname) {
        print "Updating field $field in table $table ...\n";
        my $type = $$ref[1];
        $type .= " NOT NULL" if $$ref[3];
        $dbh->do("ALTER TABLE $table
                  CHANGE $field
                  $newname $type");
    }
}

sub AddField ($$$)
{
    my ($table, $field, $definition) = @_;

    my $ref = GetFieldDef($table, $field);
    return if $ref; # already added?

    print "Adding new field $field to table $table ...\n";
    $dbh->do("ALTER TABLE $table
              ADD COLUMN $field $definition");
}

sub DropField ($$)
{
    my ($table, $field) = @_;

    my $ref = GetFieldDef($table, $field);
    return unless $ref; # already dropped?

    print "Deleting unused field $field from table $table ...\n";
    $dbh->do("ALTER TABLE $table
              DROP COLUMN $field");
}

# this uses a mysql specific command. 
sub TableExists ($)
{
   my ($table) = @_;
   my @tables;
   my $dbtable;
   my $exists = 0;
   my $sth = $dbh->prepare("SHOW TABLES");
   $sth->execute;
   while ( ($dbtable) = $sth->fetchrow_array ) {
      if ($dbtable eq $table) {
         $exists = 1;
      } 
   } 
   return $exists;
}   

# If you had to change the --TABLE-- definition in any way, then add your
# differential change code *** A B O V E *** this comment.
#
# That is: if you add a new field, you first search for the first occurence
# of --TABLE-- and add your field to into the table hash. This new setting
# would be honored for every new installation. Then add your
# AddField/DropField/ChangeFieldType/RenameField code above. This would then
# be honored by everyone who updates his DCVS installation.
#
#
# Final checks...

print("\n".
      "Well, then.  Everything looks great.\n".
      "You might still need to configure Apache to run dcvs.cgi,\n".
      "and might want to run dcvsd.pl.  Enjoy!\n\n");




