#!/usr/bin/perl -w
use strict;
#use diagnostics;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

use Date::Parse;
use Digest::MD5 qw(md5_base64);

use Getopt::Long;
my($verbose,$timespan) = (1,100);
GetOptions ('verbose' => \$verbose,
            'quiet'   => sub { $verbose = 0 },
            'timespan=s'   => \$timespan );

# Add the directory of the executable to the module search path.
BEGIN { unshift @INC, "/var/www/html/dcvs"; }

use cvsdb;

#
#  Please see the rcsfile (5) man page for details on the file format.
#
#  We don't worry about the order of phrases within a section.

my $ungetParsedElem;
my $readBuffer = "";

sub UngetParsedElement
{
   defined($ungetParsedElem) and die("internal error");
   $ungetParsedElem = $_[0];
   undef;
}

#  This will get either a @-delimited string, a semicolin, or a token from the file.
#  Trans-call state is kept in $ungetParsedElem and $readBuffer
sub GetParsedElement
{
   my($fh) = @_;

   if(defined($ungetParsedElem)) {
      my $return = $ungetParsedElem;
      $ungetParsedElem = undef;
      return $return;
   }
   my $elem;
   while(1) {  # while we need to read more for our first token
      if($readBuffer =~ s/^\s*([^\s;]+)([\s;])/$2/s || $readBuffer =~ s/^\s*(;)//s) {
         $elem = $1;
         last;
      };
      my $newRead;
      if(read($fh,$newRead,100) == 0) {  # if running out of new data
         return undef if($readBuffer =~ /^\s+$/); 
         $readBuffer .= "\n";  # make sure we end file with white space
      } else {
         $readBuffer .= $newRead;
      };
   };
   return $elem if($elem =~ /^[^\@]/);  # doesn't begin with @
   $elem .= $readBuffer;
   while(1) { # while we need to read more to complete the string
      if($elem =~ s/^\@(([^\@]*@@)*[^\@]*)\@([^\@].*)/$1/s) {
         $readBuffer = $3;
         $elem =~ s/\@\@/\@/g;
         return $elem;
      };
      my $readLength = (length($elem) < 100) ? 100 : length($elem);
      my $newRead;
      if(read($fh,$newRead,$readLength) == 0) {  # if running out of new data
         die("file ended in middle of string");
      };
      $elem .= $newRead;
   }
}

sub readSection {
   my($fh,$hashref,$comment) = @_;

   my $first = GetParsedElement($fh);
   if($first =~ /^[\d\.]+$/) {
      $$hashref{'num'} = $first;
      $first = GetParsedElement($fh);
   }
   while(1) {
      my @elems;
      while(defined(my $elem = GetParsedElement($fh))) {
         last if($elem eq ";");
         push(@elems,$elem);
      };
      $$hashref{$first} = join(" ",@elems);
      $first = GetParsedElement($fh);
      if($first =~ /^[\d\.]+$/ || $first eq "desc") {
         UngetParsedElement($first);
         return $first;
      };
   }
}

# Will return undef if end of file is reached
sub readDeltaText {
   my($fh,$hashref,$comment) = @_;

   my $first = GetParsedElement($fh);
   return undef if(!defined($first));
   if($first !~ /^[\d\.]+$/) {
      die("bad num ($first) at start of deltatext");
   }
   $$hashref{'num'} = $first;
   my $log = GetParsedElement($fh);
   if($log ne "log") {
      die("expected 'log' found $log");
   };
   $log = GetParsedElement($fh);
   $$hashref{'log'} = $log;
   $first = GetParsedElement($fh);
   while(1) {
      last if($first eq "text");
      my @elems;
      while(defined(my $elem = GetParsedElement($fh))) {
         last if($elem eq ";");
         push(@elems,$elem);
      };
      $$hashref{$first} = join(" ",@elems);
      $first = GetParsedElement($fh);
   }
   $$hashref{'text'} = GetParsedElement($fh);
   return $$hashref{'num'};
}

# FIXME: what about revisions in the database not in the RCS file
sub loadRcsFile
{
   my($file_id,$filenameWithPath) = @_;

   local(*RCS);
   if(!open(RCS,"<$filenameWithPath")) {
      print("Open failed: $filenameWithPath $!\n");
      return;
   };
   my %admin;
   $admin{'file_id'} = $file_id;
   for my $id qw(head branch access symbols locks comment expand strict) {
      $admin{$id} = "";
   }
   my $mode = "u=rw,g=rw,o=r";
   $mode = "u=rwx,g=rwx,o=rx" if -X RCS;
   my $nextline = readSection(*RCS,\%admin,"admin");
   db::SendSQL("UPDATE dirent SET head = ". db::SqlQuote($admin{'head'}) . 
                      ", comment = ". db::SqlQuote($admin{'comment'}) .
                      ", expand = ". db::SqlQuote($admin{'expand'}) .
                      ", mode = ". db::SqlQuote($mode) . 
                      " WHERE file_id = $file_id\n");
   my @revnums = ();  # store revisions in a list so we can sort them by revision number
   my %revs;
   while(1) {  # for every delta section, load it
      last if $nextline eq "desc";
      my %delta;
      for my $id qw(num date author state branches next) {
         $delta{$id} = "";
      }
      $nextline = readSection(*RCS,\%delta,"delta");
      $delta{'date'} =~ s/^(\d+)\.(\d+)\.(\d+)\.(\d+)\.(\d+)\.(\d+)$/$1-$2-$3 $4:$5:$6/; # to SQL format
      $delta{'timestamp'} = ::str2time($delta{'date'},"UTC");
      if($delta{'timestamp'} > $db::now) { 
          print("File=$filenameWithPath revision=$delta{'num'} ". 
                ::time2str("%C",$delta{'timestamp'})." is in the future!\n");
      }
      push(@revnums,$delta{'num'});
      $revs{$delta{'num'}} = \%delta;
   }
   my(%prevOfBranch,%diffParentOfChild);   # for reconstructing text for md5
   @revnums = sort cvsdb::rev_cmp @revnums ;  # so rev_id is ordered nicely 
   for my $revnum (@revnums) {  # for every delta section, store in database
      my $branchnum = "";
      if($revnum =~ /^(\d+\.\d+\..*)\.\d+$/) {  # if more than two decimals
         $branchnum = $1;
         if(defined $prevOfBranch{$branchnum}) {
            $diffParentOfChild{$revnum} = $prevOfBranch{$branchnum};
         } else {
            (my $branchpoint = $branchnum) =~ s/\.[^\.]+$//;  # remove last dot and num
            $diffParentOfChild{$revnum} = $branchpoint; 
         };
      } else { # else this is on the main trunk
         $diffParentOfChild{$prevOfBranch{""}} = $revnum if defined $prevOfBranch{""};  # main trunk goes opposite direction
      };
      $prevOfBranch{$branchnum} = $revnum;
      my $deltaref = $revs{$revnum};
      db::SendSQL("REPLACE revision SET revision = ". db::SqlQuote($revnum) . 
                      ", file_id = $file_id ". 
                      ", state = ". db::SqlQuote($$deltaref{'state'}) .
                      ", branches = ". db::SqlQuote($$deltaref{'branches'}) .
                      ", timestamp = $$deltaref{'timestamp'}" .
                      ", next = ". db::SqlQuote($$deltaref{'next'}) .
                      ", md5 = ". db::SqlQuote($$deltaref{'md5'} || "") .
                      ", branchnum = ". db::SqlQuote($branchnum));
   }
   my %numDiffChildren;  # so we no if we need to save the text when doing md5
   foreach my $parent (values %diffParentOfChild) { $numDiffChildren{$parent}++ };
   GetParsedElement(*RCS);  # desc
   GetParsedElement(*RCS);  # string
   my $firstDeltaTextFlag;
   my %savedText;
   while(1) {  # for every deltatext section
      my %deltatext;
      my $num = readDeltaText(*RCS,\%deltatext,"deltatext");
      last if ! defined($num);
      if(!defined($revs{$num})) {
         print("Partial info for revision $num for $filenameWithPath\n");
         next;
      }
      db::SendSQL("SELECT comment_id, timestamp FROM comment WHERE descr = " . db::SqlQuote($deltatext{'log'}) .
              " AND author = " . db::SqlQuote($revs{$num}->{'author'}));
      my $comment_id;
      while(my($c_id,$timestamp) = db::FetchSQLData()) {
         my $diff = $revs{$num}->{'timestamp'} - $timestamp;
         if($diff < $timespan && $diff > -$timespan) {
             $comment_id = $c_id;
             last;
         };
      };
      if(!$comment_id) {
         db::SendSQL("INSERT comment SET descr = " . db::SqlQuote($deltatext{'log'}) .
                 ", author = " . db::SqlQuote($revs{$num}->{'author'}) . ", ".
                 " timestamp = " . $revs{$num}->{'timestamp'});
         $comment_id = db::last_insert_id();
      }
      my $textLinesRef;
      my $parent = $diffParentOfChild{$num};
      if(defined $parent) { # if this delta text is actually a delta
         if($numDiffChildren{$parent}-- <= 0) {    # if this is the last use of the saved text
            $textLinesRef = delete $savedText{$parent};
         } else {   # else this is not the last use so we need to make a copy
            my @lines = @{$savedText{$parent}};
            $textLinesRef = \@lines;
         } 
         if(! defined $textLinesRef) {
            die("Missing text of $parent for $num of $filenameWithPath\n");
            $textLinesRef = [];  # reference to empty list
         }
         cvsdb::applyDiffs($deltatext{'text'},$textLinesRef);
      } else {  # else the deltatext is an actual file not a delta because this revision is the head
         my @textLines = split(/^/m,$deltatext{'text'});  # The first deltatext is not a delta, but the actual text
         $textLinesRef = \@textLines;
      }
      $savedText{$num} = $textLinesRef if $numDiffChildren{$num};
      my $md5 = ::md5_base64(join("",@$textLinesRef));
      if(defined $revs{$num}{"md5"} && length $revs{$num}{"md5"} && $revs{$num}{"md5"} ne $md5) {
         print("Bad md5 sum for revision $num of $filenameWithPath\n");
      }
      db::SendSQL("UPDATE revision SET comment_id = $comment_id, text = ". db::SqlQuote($deltatext{'text'}) . 
                  " ,md5 = ". db::SqlQuote($md5) . " ,length  = ". length(join("",@$textLinesRef)) .
                  " WHERE file_id = $file_id AND revision = " . db::SqlQuote($num));
   }
   # Now we need to populate the btime table so we can select lead revisions of branches with timespecs using only SQL.
   db::SendSQL("REPLACE btime SET file_id = $file_id, branchname = 'MAIN', branchnum = '', timestamp = NULL");
   for my $pair (split(/\s+/,$admin{'symbols'})) {  # for every symbol (defining a branch name)
      my($name,$symbol) = split(/:/,$pair);
      if(!defined($symbol)) {
         print("Bad branch symbol $pair of $filenameWithPath\n");
         next;
      }
      my $file_id_and_branchname = "file_id = $file_id, branchname = " . db::SqlQuote($name) . ", ";
      my $trunk = $symbol;
      my $numPeriods  = ($trunk =~ tr/\././);   # even num of periods mean definitely a branch
      if($trunk !~ s/\.0(\.[^\.]+)$/$1/ && ($numPeriods & 1) == 1) { # delete the second to last digit which better be a zero
         db::SendSQL("REPLACE label SET file_id = $file_id, label = " . db::SqlQuote($name) . ", ".
                     " revision = ". db::SqlQuote($symbol));
         next;
      }
      db::SendSQL("REPLACE btime SET $file_id_and_branchname" .
                  "branchnum = ".db::SqlQuote($trunk));
      while($trunk =~ s/\.([^\.]+)\.[^\.]+$//) {  # if has two periods, remove them
         my $branchpoint = "$trunk.$1";
         db::SendSQL("SELECT timestamp FROM revision WHERE file_id = $file_id AND ".
                     "revision = ".db::SqlQuote($branchpoint) );
         my $timestamp = db::FetchOneColumn();
         if(!defined($timestamp)) {
            print("Missing revision $branchpoint needed for branch $name of $filenameWithPath\n");
            next;
         }
         $trunk = "" if $trunk =~ /^[^\.]$/;  # if no periods (thus main trunk special case)
         db::SendSQL("REPLACE btime SET $file_id_and_branchname".
                     "branchnum = ".db::SqlQuote($trunk) . ", timestamp = $timestamp"); 
      }
   };
   close(RCS);
}   

sub cleanPath {
   my($path) = @_;

   $path =~ s|/+|/|g;  # eliminate repeated slashes
   while($path =~ s|/[^/]+/\.\./|/|g) {};  # eliminate parent dir references /../
   while($path =~ s|/\./|/|g) {};  # eliminate self dir references /./
   #$path =~ s|./$||g;  # eliminate trailing slash
   return $path;
}

sub loadRcsRecurse {
   my ($dbPathId,$diskPath) = @_;

   $diskPath =~ /([^\/]+)$/;  # get leaf-most dir or filename
   my $name = $1;
   if($name eq "CVS") {
      print("Can't version control a file or directory named CVS.\n");
   } elsif(-d $diskPath && $name eq "Attic") {
      local(*DIR);
      opendir(DIR,"$diskPath") or die;
      readdir(DIR); readdir(DIR);
      my @dirents = readdir(DIR);
      close(DIR);
      for my $file (@dirents) {
         loadRcsRecurse($dbPathId,"$diskPath/$file");
      }
   } elsif(-d $diskPath)  {  # if is a directory
      print("loading RCS files from $diskPath\n");
      db::SendSQL("SELECT file_id,type FROM dirent WHERE fname = ". db::SqlQuote($name) . " AND ".
                  "parent = $dbPathId"); 
      my($file_id,$type) = db::FetchSQLData();
      if(! defined $file_id) {
         $file_id = cvsdb::createDir($name,$dbPathId);
      } elsif($type ne 'dir') {
         print("Directory $diskPath skipped because $type of same name exists.\n");
         next;
      };
      local(*DIR);
      opendir(DIR,"$diskPath") or die;
      readdir(DIR); readdir(DIR);
      my @dirents = readdir(DIR);
      close(DIR);
      for my $file (@dirents) {  # do non directories first 
         loadRcsRecurse($file_id,"$diskPath/$file") if ! -d "$diskPath/$file";
      }
      for my $file (@dirents) {
         loadRcsRecurse($file_id,"$diskPath/$file") if -d "$diskPath/$file";
      }
   } elsif(! -e $diskPath) {
      print("File $diskPath does not exist.\n");
   } elsif(! -r $diskPath) {
      print("Cannot read $diskPath\n");
   } elsif($name =~ s/,v$//) {
      my $params = "fname = ". db::SqlQuote($name) . ", type = 'file', parent = $dbPathId";
      db::SendSQL("SELECT file_id FROM dirent WHERE fname = ". db::SqlQuote($name) ." AND parent = $dbPathId"); 
      my $file_id = db::FetchOneColumn();
      if(defined($file_id)) {
         db::SendSQL("UPDATE dirent SET $params WHERE file_id = $file_id"); 
      } else {
         db::SendSQL("INSERT dirent SET $params"); 
         db::SendSQL("SELECT LAST_INSERT_ID()"); 
         $file_id = db::FetchOneColumn();
      };
      loadRcsFile($file_id,$diskPath);
   } else {
      print("File $diskPath not a RCS file.\n");
   };
}

# Trailing slash on directories is relevant.
sub loadRcs {
   my ($dbPath,$diskPath) = @_;

   $|++;  # disable output buffering
   $dbPath = "/" if ! defined($dbPath);
   my $dbPathId = cvsdb::id($dbPath);
   return if $dbPathId == -1;
   if(! -e $diskPath) {
      print("Local filesystem path $diskPath does not exist.\n");
      return;
   }
   db::getDbNowTime();
   db::SendSQL("LOCK TABLES dirent WRITE, revision WRITE, btime WRITE, comment WRITE, label WRITE, descendent WRITE");
   if($diskPath =~ /\/$/) {  # if end with a slash
      if(! -d $diskPath) {
         print("Local filesystem directory $diskPath does not exist.\n");
         return;
      }
      local(*DIR);
      opendir(DIR,"$diskPath") or die; 
      readdir(DIR); readdir(DIR);  #   . and ..
      for my $file (readdir(DIR)) {
         loadRcsRecurse($dbPathId,"$diskPath/$file");
      }
      closedir(DIR);
   } else {
      loadRcsRecurse($dbPathId,$diskPath);
   };
   db::SendSQL("UNLOCK TABLES");
   $|--;  # possibly re-enable output buffering
}

if($#ARGV < 1) {  # if one command line parameter or fewer
   print("usage: loadrcs.pl [ options ] dbPath diskPath ..\n".
         "    --timespan=100 (max delta in seconds for two same-comment commits to be the same)\n" );
   exit 1;
}
my $dbPath = shift @ARGV;
foreach my $diskPath (@ARGV) {
   $diskPath =~ s/\/+$//;  # trim trailing slash
   loadRcs($dbPath,$diskPath);
}

1;

