#!/usr/bin/perl -w -T
use strict;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

use diff;

package merge;

my $debug = undef;

#  This will dammage the @match and @replace.
sub convertToCDiff
{
   my($oldFirstline,$newFirstline,$matchRef,$replaceRef) = @_;

   my $changes = diff::diff($matchRef,$replaceRef);
   my $cdiff = "@@ -".($oldFirstline+1).",".($#$matchRef+1)." +".($newFirstline+1).",".($#$replaceRef+1)." @@\n";
   foreach my $hunk (@$changes) {  #  for each diff::diff() hunk (not unidiff hunk)
      foreach my $linechange (@$hunk) {  # for each individual line addition or subtraction
         my($plusminus,$offset,$text) = @$linechange;
         if($plusminus eq "+") {
            while($text ne $$replaceRef[0]) { 
               $cdiff .= " " .  shift(@$replaceRef);  # transfer common line
               shift(@$matchRef);
            };
            $cdiff .= "+" . shift(@$replaceRef);  # transfer added line
         } else {
            while($text ne $$matchRef[0] && $#$matchRef > -1) {
               $cdiff .= " " .  shift(@$replaceRef);  # transfer common line
               shift(@$matchRef);
            };
            $cdiff .= "-" . shift(@$matchRef);  # transfer removed line
         };     
      };
   };
   $cdiff .= " " . join(" ",@$replaceRef);  # transfer trailing context
   return $cdiff;
}

sub tryMatch
{
   my ($matchOffset,$matchRef,$replaceRef,$targetRef,$frontCheat,$backCheat) = @_;

   # shame perl 5 doesn't have a list equality operator.
   foreach my $index (0+$frontCheat .. $#$matchRef-$backCheat) {
      return 0 unless $$matchRef[$index] eq $$targetRef[$matchOffset + $index];  # return "non-match"
   }
#print("tryMatch() found match! $matchOffset\n");
   splice(@$targetRef,$matchOffset,$#$matchRef+1,@$replaceRef);
   return 1;
}

#  Try the match at different places
sub tryMatchSearch 
{
   my ($matchOffset,$matchRef,$replaceRef,$targetRef,$fc,$bc) = @_;
   print("match> ",join("match> ",@$matchRef),"\n") if $debug;
   print("replace> ",join("replace> ",@$replaceRef),"\n") if $debug;
   #  First we check the actual location, then offset -1,1,-2,2,-3,3, etc.
   if(tryMatch($matchOffset,$matchRef,$replaceRef,$targetRef,$fc,$bc)) {
      return 1;
   };
   for(my $radius=1;$matchOffset >= $radius || $matchOffset + $radius + $#$matchRef < $#$targetRef;$radius++) {
      my $start = $matchOffset-$radius;
      if($start >= 0) {
         if(tryMatch($start,$matchRef,$replaceRef,$targetRef,$fc,$bc)) {
            return 1;
         }
      }
      $start = $matchOffset+$radius;
      if($start + $#$matchRef <= $#$targetRef) {
         if(tryMatch($start,$matchRef,$replaceRef,$targetRef,$fc,$bc)) {
            return 1;
         }
      }
   }
   return 0;
}

#  First try with three lines of context, and if it doesn't work try two.
sub doChange 
{
   my ($matchOffset,$matchRef,$replaceRef,$targetRef) = @_;

   my($frontCheat,$backCheat) = (0,0);  # how man lines of context to ignore
   return 1 if tryMatchSearch($matchOffset,$matchRef,$replaceRef,$targetRef,$frontCheat,$backCheat);   
   if($$matchRef[0] eq $$replaceRef[0] &&
      $$matchRef[1] eq $$replaceRef[1] &&
      $$matchRef[2] eq $$replaceRef[2]) {  # if first three lines of context are same
      $frontCheat++;
   }
   if($$matchRef[-1] eq $$replaceRef[-1] &&
      $$matchRef[-2] eq $$replaceRef[-2] &&
      $$matchRef[-3] eq $$replaceRef[-3]) {  # if last three lines of context are same
      $backCheat++;
   }
   return 1 if tryMatchSearch($matchOffset,$matchRef,$replaceRef,$targetRef,$frontCheat,$backCheat);   
   return 0;
}

#  Given one old version and two new ones, we try to merge the changes.
#  We literally create a context diff and patch with it.  
#  Rejects are provided as a context diff.
#
#  To start with we are simply going to require a perfect match for
#  all the context of a cdiff of 3 line context.  This means if the
#  diff::diff hunks have less than three lines inbetween them, we
#  will consider then one hunk.
#  For the deletes we will make sure all the deleted lines match.
#
#  Unlike the output of GNU diff, here the first line is number zero.
#
sub merge {
   my($oldtext,$targetText,$newText) = (@_);

   my @target = split(/^/,$targetText);
   my @old = split(/^/,$oldtext);
   my @new = split(/^/,$newText);
   my $change = diff::diff(\@old,\@new);

   # For each hunk.  (not diff:diff() hunk but unified-context-diff hunk)
   # we compose a @matchlines (containing context) and @newlines.
   # The terms "old" and "new" generally refer to the two files.
   my($oldFirstline,$oldLastline);  # for context to match  (The OLDer file)
   my($newFirstline,$newLastline);  # includes context too  (The NEWer file)
   my $newGrowth = 0;   #  add to old line index to get new
   my $targetGrowth = 0;   #  add to target line index to get post-patch locations (hopefully hunks apply in order)
   my $reject = "";  # here we accumulated rejected hunks as a cdiff
   foreach my $hunk (@$change,[["end",2000000000,"\n"]]) {  #  for each diff::diff() hunk (not unidiff hunk)
      foreach my $linechange (@$hunk) {  # for each individual line addition or subtraction
         my($plusminus,$offset,$text) = @$linechange;
         print("$plusminus offset=$offset $text") if $debug;
         my $oldOffset = $plusminus eq "-" ? $offset : $offset - $newGrowth;
         my $newOffset = $plusminus eq "+" ? $offset : $offset + $newGrowth;
         if(!defined $oldFirstline) {  # if this was the first change of the unidiff hunk
            $oldFirstline = $oldOffset - 3;
            $oldFirstline = 0 if $oldFirstline < 0;
            $oldLastline = $oldOffset-1;
            $newFirstline = $newOffset - 3;
            $newFirstline = 0 if $newFirstline < 0;
            $newLastline = $newOffset-1;
         } elsif($oldLastline + 5 < $oldOffset) {  # if this new line change is not the current hunk (truely should be +6)
            print("# oldFirstl=$oldFirstline oldLastl=$oldLastline newFirstl=$newFirstline newLastl=$newLastline\n") if $debug;
            $oldLastline += 3;  # create trailing context
            $newLastline += 3;
            $oldLastline = $#old if $oldLastline > $#old;  
            $newLastline = $#new if $newLastline > $#new;
            my @match = @old[$oldFirstline..$oldLastline];  # strangely, in perl5 ref to slice gives ref to only first scalar
            my @replace = @new[$newFirstline..$newLastline];
            if(doChange($oldFirstline+$targetGrowth,\@match,\@replace,\@target)) {  # if hunk succeeded
               $targetGrowth += $#replace - $#match;
            } else {  # else we have a conflict
               $reject .= convertToCDiff($oldFirstline,$newFirstline,\@match,\@replace);
            }
            last if $plusminus eq "end";
            $oldFirstline = undef;
            redo;
         };
         last if $plusminus eq "end";
         if($plusminus eq "-") {
             $newGrowth--;
             $oldLastline++;
         } else {
             $newGrowth++;
             $newLastline++;
         };
      };
   };
   return (join("",@target),$reject);  
}


1;

__END__


   For diff::diff()...

         a b c e h j l m n p
         b c d e f j k l m r s t

       Result:

        [
          [ [ '-', 0, 'a' ] ],

          [ [ '+', 2, 'd' ] ],

          [ [ '-', 4, 'h' ] ,
            [ '+', 4, 'f' ] ],

          [ [ '+', 6, 'k' ] ],

          [ [ '-', 8, 'n' ],
            [ '-', 9, 'p' ],
            [ '+', 9, 'r' ],
            [ '+', 10, 's' ],
            [ '+', 11, 't' ],
          ]
        ]


