#!/usr/bin/perl -w -T
use strict;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>


#  An SQL guru could probably do most of this processing in SQL
#  and the script might run a lot faster.

#  None of these errors should be found if the database, perl, and
#  the DCVS code are bugfree.

#  We check some things which are disallowed by the database schema,
#  because the schema can have bugs too.

use Getopt::Long;
my($verbose,$fix) = (1," (fixed)"); 
GetOptions ('verbose' => \$verbose,
            'quiet'   => sub { $verbose = 0 },
            'fix'     => sub { $fix = " (fixed)" },
            'nofix'   => sub { $fix = "" });

# Add the directory of the executable to the module search path.
BEGIN { unshift @INC, "/var/www/html/dcvs"; }

use config;
use db;
use cvsdb;

my $ansiBright = "\033[1m";
my $ansiBlue = "\033[34m";
my $ansiGreen = "\033[32m";
my $ansiRed = "\033[31m";
my $ansiNormal = "\033[0m";

if($fix) {
   db::SendSQL("LOCK TABLES dirent WRITE, revision WRITE, btime WRITE, comment WRITE, label WRITE, ".
               "descendent WRITE, groups WRITE, user WRITE");
} else {  # else don't fix
   db::SendSQL("LOCK TABLES dirent READ, revision READ, btime READ, comment READ, label READ, ".
               "descendent READ, groups READ, user READ");
}

my %direntTypeCount; # key is type
my(%parentOfFile,%parentOfDir,%fname);  # key is file_id
my %dirChild;  # key is "$parent-$fname"

print "Loading dirent table.\n" if $verbose;
db::SendSQL("SELECT file_id, parent, fname, type, path FROM dirent");
while(my($file_id,$parent,$fname,$type,$path) = db::FetchSQLData()) {
   if($fname eq "" && $file_id != 0) {
      print("${ansiRed}Found empty filename ($file_id)$ansiNormal\n");
   } elsif($fname eq "." || $fname eq "..") {
      print("${ansiRed}Found illegal filename $fname ($file_id)$ansiNormal\n");
   } elsif($fname =~ /[\/\0]/) {
      print("${ansiRed}Found illegal char in filename $fname ($file_id)$ansiNormal\n");
   }
   if($file_id == 0) {
      if($fname ne "" || $parent != -1 || $type ne 'dir' || $path ne '/') {
         print("${ansiRed}Corrupted root directory record (file_id=0).$ansiNormal\n");
      }
   } elsif($file_id == -1) {  # -1 is used as error code
      print("${ansiRed}Dirent table has record with file_id of $file_id ($fname).$ansiNormal\n");
   }
   if(exists $fname{$file_id}) {
      print("${ansiRed}Duplicate file_id $fname and $fname{$file_id} ($file_id)$ansiNormal\n");
   }
   $fname{$file_id} = $fname;
   $direntTypeCount{$type}++;
   if($type eq 'dir') {
      $parentOfDir{$file_id} = $parent;
   } else {
      $parentOfFile{$file_id} = $parent;
   }
   if(defined $dirChild{"$parent-$fname"}) {
      print("${ansiRed}Duplicate filename $fname in same directory $dirChild{$file_id}. $ansiNormal\n");
   }
   $dirChild{"$parent-$fname"} = $file_id;
}
%dirChild = ();  # free some memory.
print "Counting types.\n" if $verbose;
foreach my $type (sort keys %direntTypeCount) {
   my $error = "";
   $error = "${ansiRed}ERROR$ansiNormal" if $type ne "file" && $type ne "dir";
   print("   $direntTypeCount{$type} \"$type\" $error\n");
}
print "Finding directory loops.\n" if $verbose;
my $loops;
foreach my $dir (sort keys %parentOfDir) {  # for each directory
   next if $dir == 0;
   my %seen;
   my $parent = $dir;
   my($error,$error2) = ("","");
   while(1) {  # while we trace this directorie's ancestry
      $error = "$fname{$parent}/$error";
      $error2 = "$parent/$error2";
      if($seen{$parent}) {
         print("${ansiRed}Found directory loop ..../$error (..../$error2)$ansiNormal\n");
         $loops++;
         last;
      }
      $seen{$parent}++;
      $parent = $parentOfDir{$parent};
      if(! defined $parent) {  
         print("${ansiRed}Found parent nonexistent ????$error ($error2)$ansiNormal\n");
         $loops++;
         last;
      }
      last if $parent == 0;  # root is number 0
   }
}
print "File parents must be directories.\n" if $verbose;
foreach my $file (sort keys %parentOfFile) {  # for each directory
   my $parent = $parentOfFile{$file};
   if(defined $parentOfFile{$parent}) {
      print("${ansiRed}File $fname{$file} ($file) has parent \"$fname{$parent}\" ($parent) of type file.$ansiNormal\n"); 
      next;
   }
   if($parent != 0 && !defined $parentOfDir{$parent}) {
      print("${ansiRed}File $fname{$file} ($file) has nonexistent parent ($parentOfFile{$file}).$ansiNormal\n"); 
      next;
   }
}

my %descendent;

print "Loading descendent table.\n" if $verbose;
db::SendSQL("SELECT parent, child FROM descendent");
my @sql;
while(my($parent,$child) = db::FetchSQLData()) {
   if(defined $parentOfFile{$parent}) {
      print("${ansiRed}Descendent table has file $fname{$parent} ($parent) as parent (of $child).$ansiNormal$fix\n"); 
      push(@sql,"DELETE FROM descendent WHERE parent = $parent AND child = $child") if $fix;
   } elsif($parent == 0) {
      print("${ansiRed}Descendent table has a record with a parent of zero.$ansiNormal$fix\n"); 
      push(@sql,"DELETE FROM descendent WHERE parent = $parent AND child = $child") if $fix;
   } elsif(! defined $parentOfDir{$parent}) {
      print("${ansiRed}Descendent table has a parent ($parent) not in dirent table.$ansiNormal$fix\n"); 
      push(@sql,"DELETE FROM descendent WHERE parent = $parent AND child = $child") if $fix;
   }
   if(defined $parentOfFile{$child}) {
      print("${ansiRed}Descendent table has file $fname{$child} ($child) as child (of $parent).$ansiNormal$fix\n"); 
      push(@sql,"DELETE FROM descendent WHERE parent = $parent AND child = $child") if $fix;
   } elsif(! defined $parentOfDir{$child}) {
      print("${ansiRed}Descendent table has a child ($child) not in dirent table.$ansiNormal$fix\n"); 
      push(@sql,"DELETE FROM descendent WHERE parent = $parent AND child = $child") if $fix;
   }
   my $packed = pack("LL",$parent,$child); 
   if($descendent{$packed}) {
      print("${ansiRed}Duplicate descendent???? parent=$parent child=$child.$ansiNormal\n"); 
   }
   $descendent{$packed} = 1; 
}
foreach (@sql) { db::SendSQL($_) };
@sql = ();
if($loops) {
   print("${ansiRed}Skipping descendent check because of looping parentage in dirent table.$ansiNormal\n"); 
} else {
   print "Recomputing lineages.\n" if $verbose;
   foreach my $childDir (keys %parentOfDir) {
      next if $childDir == 0;  # root dir has bogus parent
      my $parent = $parentOfDir{$childDir};
      while($parent) {
         my $packed = pack("LL",$parent,$childDir); 
         if(! $descendent{$packed}) {
             print("${ansiRed}Missing descendent rule where parent=$parent child=$childDir.$ansiNormal$fix\n"); 
             push(@sql,"INSERT descendent SET parent = $parent AND child = $childDir") if $fix;
         }          
         $parent = $parentOfDir{$parent};
      }
   }
}
foreach (@sql) { db::SendSQL($_) };
@sql = ();

sub removeFile
{
   my($file_id) = @_;

   db::SendSQL("DELETE FROM dirent WHERE file_id = $file_id");
   db::SendSQL("DELETE FROM btime WHERE file_id = $file_id");
   db::SendSQL("DELETE FROM label WHERE file_id = $file_id");
   db::SendSQL("DELETE FROM revision WHERE file_id = $file_id");
   db::SendSQL("DELETE FROM descendent WHERE parent = $file_id");
   db::SendSQL("DELETE FROM descendent WHERE child = $file_id");
}

my(%revOfRevId);  # key is rev_id
my(%numRevsOfFile,%numMainTrunkRevsOfFile);  # key is file_id
my(%timestampOfRev,%rev_idOfRev);  # key is $rev.  They are reset every file
my $prevFile_id = -1;

print "Loading revision table.\n" if $verbose;
db::SendSQL("SELECT file_id, rev_id, revision, comment_id, state, branches, branchnum, next, timestamp ".
            "FROM revision ORDER BY file_id, rev_id");
while(my($file_id,$rev_id,$rev,$comment_id,$state,$branches,$branchnum,$next,$timestamp) = db::FetchSQLData()) {
   if($file_id != $prevFile_id) {
      %timestampOfRev = %rev_idOfRev = ();
      $prevFile_id = $file_id;
   }
   if(! defined $parentOfFile{$file_id}) {
      print("${ansiRed}Revision $rev ($rev_id) has nonexistent file ($file_id).$ansiNormal\n"); 
   }
   if(defined $revOfRevId{$rev_id}) {
      print("${ansiRed}Duplicate rev_id ($rev_id) $rev and $revOfRevId{$rev_id} of $fname{$file_id} ($file_id).$ansiNormal\n"); 
      next;
   }
   if($rev eq "") {
      print("${ansiRed}Empty revision name ($rev_id) of $fname{$file_id} ($file_id).$ansiNormal\n"); 
   }
   if($rev !~ /^[\d\.]*$/) {
      print("${ansiRed}Illegal revision name character $rev ($rev_id) of $fname{$file_id} ($file_id).$ansiNormal\n"); 
   }
   if($rev !~ /\./) { 
      print("${ansiRed}Revision names must have at least one dot. $rev ($rev_id) of $fname{$file_id} ($file_id).$ansiNormal\n"); 
   }
   if($state ne "Exp" && $state ne "dead") {
      print("${ansiRed}Illegal state $state for $rev ($rev_id) of $fname{$file_id} ($file_id).$ansiNormal\n"); 
   }
   (my $rightBranchnum = $rev) =~ s/\.[^\.]+$//;  # delete last dot and trailing non-dot stuff
   $rightBranchnum = "" if $rightBranchnum !~ /\./; # use blank if no dot is present
   if($rightBranchnum ne $branchnum) { 
      print("${ansiRed}Revision $rev of $fname{$file_id} ($file_id) has wrong branchnum '$branchnum' not '$rightBranchnum'.$ansiNormal$fix\n"); 
      db::SendSQL("UPDATE revision SET branchnum = ".db::SqlQuote($rightBranchnum)." WHERE rev_id = ".$rev_id) if $fix;
   }
   if(! defined $timestamp) {
      print("${ansiRed}Revision $rev of $fname{$file_id} ($file_id) has undefined timestamp.$ansiNormal\n"); 
   } else {
      # In cvsdb.pm we select based on rev_id with the assumption that timestamps are in that order.
      my $prev = cvsdb::GuessPreviousRev($rev);
      # for every conceivable ancestor until we find one
      for(my $prev = cvsdb::GuessPreviousRev($rev);length $prev;$prev = cvsdb::GuessPreviousRev($prev)) {  
         next if ! defined $timestampOfRev{$prev};  
         if($timestampOfRev{$prev} > $timestamp) {
            print("${ansiRed}Revision $rev of $fname{$file_id} ($file_id) has out of order timestamp.$ansiNormal\n"); 
         };
         if($rev_idOfRev{$prev} >= $rev_id) {
            print("${ansiRed}Revision $rev of $fname{$file_id} ($file_id) has out of order rev_id.$ansiNormal\n"); 
         };
         last; 
      }
      $timestampOfRev{$rev} = $timestamp;
      $rev_idOfRev{$rev} = $rev_id;
   };
   $revOfRevId{$rev_id} = $rev;
   $numRevsOfFile{$file_id}++;
   $numMainTrunkRevsOfFile{$file_id}++ if ($rev =~ tr/././) == 1;  # exactly one dot
}
print "Checking that each file has at least one revision.\n" if $verbose;
foreach my $file_id (keys %parentOfFile) {
   if(! $numRevsOfFile{$file_id}) {
      print("${ansiRed}File $fname{$file_id} ($file_id) has no revisions.$ansiNormal$fix\n");
      removeFile($file_id) if $fix;
   } elsif(! $numMainTrunkRevsOfFile{$file_id}) {
      print("${ansiRed}File $fname{$file_id} ($file_id) has no main trunk revisions.$ansiNormal\n");
   };
}
%numRevsOfFile = %numMainTrunkRevsOfFile = ();  # We no longer need these.

my %groupname;  # key is group_id

print "Loading groups table.\n" if $verbose;
db::SendSQL("SELECT group_id, name FROM groups");
while(my($group_id,$name) = db::FetchSQLData()){
   if(! defined $name or $name eq "") {
      print("${ansiRed}Group name for $group_id is empty.$ansiNormal\n"); 
   }
   if($group_id > 0xffff || $group_id < 0) {
      print("${ansiRed}Group $name has group_id ($group_id) out of range.$ansiNormal\n"); 
   }
   $groupname{$group_id} = $name;
}
print "Loading user's groups.\n" if $verbose;
db::SendSQL("SELECT username, groups FROM user");
while(my($username,$groups) = db::FetchSQLData()) {
   if(length($groups) & 1) {
      print("${ansiRed}Groups blobk for $username had odd length.$ansiNormal\n"); 
   }
   my(%complained,%member);  # So we don't flood with complaints.  Key is $group_id.
   foreach my $group_id (unpack("n*",$groups)) {
      if(! defined $groupname{$group_id} && ! $complained{$group_id}) {
         $complained{$group_id} = 1;
         print("${ansiRed}User $username belongs to group $group_id which is not in groups table.$ansiNormal\n");
      };
      if(defined $member{$group_id}) {
         print("${ansiRed}User $username has group $groupname{$group_id} ($group_id) listed twice.$ansiNormal\n");
      }
      $member{$group_id} = 1;
   };
}

print "Loading btime table (used for fetching branch revisions).\n" if $verbose;
my $prevFileId = -1;
my $hasMain = 1;
my $prevBranchname;
my $prevBranchnum;
db::SendSQL("SELECT file_id, branchnum, branchname, timestamp FROM btime ORDER BY file_id, branchname, -branchnum ");
REC: while(my($file_id,$branchnum,$branchname,$timestamp) = db::FetchSQLData()) {
#printf("### file_id=$file_id branchname=$branchname branchnum=$branchnum timestamp=%s\n",defined($timestamp) ? $timestamp : "undef");
   db::PushGlobalSQLState();
   if(! exists $fname{$file_id}) {
       print("${ansiRed}btime has file_id = $file_id which doesn't exist (name=$branchname).$ansiNormal$fix\n");
       db::SendSQL("DELETE FROM btime WHERE file_id = $file_id");
       next;
   };
   # Order from the database is important.  First handle end of last file_id.
   if($file_id != $prevFileId) {
      if(!$hasMain) {
         print("${ansiRed}file $fname{$prevFileId} ($prevFileId) is missing MAIN branch in btime.$ansiNormal$fix\n");
         db::SendSQL("REPLACE btime SET file_id = $prevFileId, branchname = 'MAIN', branchnum = '', timestamp = NULL");
      }
      $hasMain = 0;
   };
   $hasMain = 1 if $branchname eq "MAIN";
   # Now we need to check/fix the trunk records. (They have non-null timestamp.)
   # We fudge $prevBranchnum for this purpose.
   while(defined($prevBranchnum) && $prevBranchnum ne "" && $prevBranchname ne "MAIN"){# while we needed additional trunk records
      my @elem = split(/\./,$prevBranchnum);
      pop(@elem); pop(@elem);
      my $rightBranchnum = "";
      $rightBranchnum = join(".",@elem) if($#elem > 1);  #  if X.Y.Z or X.Y.Z.A.B or etc.
#print("----prev=$prevBranchnum right=$rightBranchnum curr-branchnum=$branchnum  $branchname-$prevBranchname $file_id-$prevFileId\n");
      last if($branchnum eq $rightBranchnum && $branchname eq $prevBranchname && $file_id == $prevFileId); # if correct
      # If the current record is same branch as the last, but has longer or wrong prefix branchnum....
      if($branchname eq $prevBranchname && $file_id == $prevFileId && (length($branchnum) >= length($rightBranchnum) ||
         $branchnum ne substr($rightBranchnum,0,length($branchnum)) )) {
         print("${ansiRed}file $fname{$file_id} ($file_id) has bad trunk branchnum '$branchnum' ".
            "(branchname=$branchname).$ansiNormal$fix\n"); 
         db::SendSQL("DELETE FROM btime WHERE file_id = $file_id AND branchname = ".db::SqlQuote($branchname).
                     " AND branchnum = ".db::SqlQuote($branchnum)) if $fix;
         next REC;
      }
      print("${ansiRed}file $fname{$prevFileId} ($prevFileId) is missing btime record ".
            "branchname=$prevBranchname branchnum='$rightBranchnum'.$ansiNormal");  # no newline
      if($fix) {
         (my $bpoint = $prevBranchnum) =~ s/\.[^\.]+$//;  # remove branchnum  X.Y.Z -> X.Y
         db::SendSQL("SELECT timestamp FROM revision WHERE file_id = $prevFileId AND revision = ".db::SqlQuote($bpoint));
         my $btime = db::FetchOneColumn();
         if(defined $btime) {
            db::SendSQL("INSERT btime SET file_id = $prevFileId, branchnum = ".db::SqlQuote($rightBranchnum).
                        ", branchname = ".db::SqlQuote($prevBranchname).", timestamp = $btime");
            print("$fix\n");
         } else {
            print("\n${ansiRed}file $fname{$prevFileId} ($prevFileId) is missing revision ".
                  "'$bpoint' needed for branchname=$prevBranchname.$ansiNormal\n");
         }
      } else {
         print("\n");
      }
      $prevBranchnum=$rightBranchnum;
   }
   # Now we are done dealing with the previous record
   if($branchname eq "MAIN") {
      if(defined $timestamp) {
         print("${ansiRed}file $fname{$file_id} ($file_id) MAIN must have undef timestamp.$ansiNormal$fix\n");
         db::SendSQL("REPLACE btime SET file_id = $file_id, branchname = 'MAIN', branchnum = '', timestamp = NULL") if $fix;
      } elsif(! defined $branchnum || $branchnum ne "") {
         print("${ansiRed}file $fname{$file_id} ($file_id) MAIN must have empty branchnum.$ansiNormal$fix\n");
         db::SendSQL("DELETE FROM btime WHERE file_id = $file_id AND branchname = 'MAIN' AND branchnum = ".db::SqlQuote($branchnum)) if $fix;
         db::SendSQL("REPLACE btime SET file_id = $file_id, branchname = 'MAIN', branchnum = '', timestamp = NULL") if $fix;
      };
   } else {  # else we are not on branch MAIN
      # The first record of a branch for a file should have null timestamp.
      if(defined $prevBranchname && ($branchname ne $prevBranchname || $file_id != $prevFileId) && defined $timestamp) {
         print("${ansiRed}file $fname{$file_id} ($file_id) has branch $branchname without null timestamp.$ansiNormal\n");
      }
   }
   if(! defined $branchnum) {
      print("${ansiRed}file $fname{$file_id} ($file_id) has undefined branchnum.$ansiNormal$fix\n");
      db::SendSQL("DELETE FROM btime WHERE file_id = $file_id AND branchnum IS NULL") if $fix;
      next;
   } elsif($branchnum ne "") {
      my $numDots = ($branchnum =~ tr/././);
      if(($numDots & 1) || $numDots == 0) {  # if odd number of dots, or no dots
         print("${ansiRed}file $fname{$file_id} ($file_id) has illegal branchnum ($branchnum) (name=$branchname).$ansiNormal$fix\n");
         db::SendSQL("DELETE FROM btime WHERE file_id = $file_id AND branchnum = ".db::SqlQuote($branchnum)) if $fix;
         next;
      };
   }
   $prevFileId = $file_id;  # This should not be in the continue block because of the next() on deleted records.
   $prevBranchnum = $branchnum;
   $prevBranchname = $branchname;
} continue {
   db::PopGlobalSQLState();
};



